#include "fetchManager.hpp"

#include "client.hpp"

#include <gflags/gflags.h>

DEFINE_uint64(max_incoming_buffered_size, 1024 * 1024 * 200,
    "maximum number of buffered incoming bytes (pending write to storage) per "
    "link");

namespace rtransfer {

namespace {
struct Header {
    std::uint64_t reqId;
    std::uint8_t priority;
    std::uint64_t offset;

    Header(std::unique_ptr<folly::IOBuf> &buf)
    {
        buf->gather(21);
        std::memcpy(&reqId, buf->data(), sizeof(reqId));
        buf->trimStart(sizeof(reqId));
        std::memcpy(&priority, buf->data(), sizeof(priority));
        buf->trimStart(sizeof(priority));
        std::memcpy(&offset, buf->data(), sizeof(offset));
        buf->trimStart(sizeof(offset) + 4);

        reqId = folly::Endian::big(reqId);
        offset = folly::Endian::big(offset);
    }
};
}  // namespace

FetchManager::FetchManager(
    StoragesMap &storages, HandleCache &handleCache, folly::EventBase *base)
    : storages_{storages}
    , handleCache_{handleCache}
    , base_{base}
{
}

folly::Future<std::size_t> FetchManager::newFetch(std::uint64_t requestId,
    folly::fbstring destStorageId, folly::fbstring destFileId,
    std::uint64_t offset, std::size_t size,
    folly::Function<void(std::uint64_t, std::size_t)> notifyCb)
{
    VLOG(1) << "New fetch: " << requestId << " " << offset << " " << size;

    auto result = fetchInfo_.emplace(
        requestId, FetchInfo{std::move(destStorageId), std::move(destFileId),
                       offset, size, {}, 0, std::move(notifyCb)});

    assert(result.second);
    ONE_METRIC_COUNTER_INC("fetches.concurrent");
    return result.first->second.promise.getFuture();
}

void FetchManager::read(Context *ctx, std::unique_ptr<folly::IOBuf> buf)
{
    base_->runInEventBaseThread(
        [this, buf = std::move(buf)]() mutable { doRead(std::move(buf)); });
}

void FetchManager::doRead(std::unique_ptr<folly::IOBuf> buf)
{
    Header header{buf};
    VLOG(2) << "Received data about " << header.reqId;

    auto it = fetchInfo_.find(header.reqId);
    if (it == fetchInfo_.end()) {
        LOG(INFO) << "Discarding canceled data for reqId = " << header.reqId;
        return;
    }

    auto &req = it->second;
    auto toWrite = buf->computeChainDataLength();

    auto storageIt = storages_.find(req.storageId);
    if (storageIt == storages_.cend())
        return cancelFetch(
            it, "no such destination storage " + folly::hexlify(req.storageId));

    noteInProgress(toWrite);
    folly::IOBufQueue queue{folly::IOBufQueue::cacheChainLength()};
    queue.append(std::move(buf));
    handleCache_
        .write(*storageIt->second, req.fileId, header.offset, std::move(queue),
            header.priority)
        .via(base_)
        .then([this, offset = header.offset, toWrite, reqId = header.reqId] {
            auto it = fetchInfo_.find(reqId);
            if (it == fetchInfo_.end())
                return;

            it->second.notifyCb(offset, toWrite);
            it->second.wrote += toWrite;
            if (it->second.wrote >= it->second.size)
                fulfillPromise(it);
        })
        .onError([this, reqId = header.reqId](const std::exception &e) {
            auto it = fetchInfo_.find(reqId);
            if (it != fetchInfo_.end())
                cancelFetch(it, e.what());
        })
        .ensure([this, toWrite] { noteDone(toWrite); });
}

void FetchManager::cancelFetch(std::uint64_t requestId)
{
    auto it = fetchInfo_.find(requestId);
    if (it == fetchInfo_.end()) {
        LOG(WARNING) << "Unable to cancel unknown fetch request " << requestId;
        return;
    }

    cancelFetch(it, "canceled");
}

void FetchManager::setTotalSize(std::uint64_t reqId, std::size_t size)
{
    auto it = fetchInfo_.find(reqId);
    if (it == fetchInfo_.end()) {
        LOG(WARNING) << "Unable to set size for unknown fetch request "
                     << reqId;
        return;
    }

    if (it->second.size < size) {
        LOG(WARNING) << "Refusing to set size to larger than requested";
        return;
    }

    it->second.size = size;
    if (it->second.size <= it->second.wrote)
        fulfillPromise(it);
}

void FetchManager::cancelFetch(
    std::unordered_map<std::uint64_t, FetchInfo>::iterator it,
    folly::StringPiece reason)
{
    ONE_METRIC_COUNTER_DEC("fetches.concurrent");
    ONE_METRIC_COUNTER_INC("fetches.cancelled");
    it->second.promise.setException(
        folly::make_exception_wrapper<std::runtime_error>(reason.str()));
    fetchInfo_.erase(it);
}

void FetchManager::fulfillPromise(
    std::unordered_map<std::uint64_t, FetchInfo>::iterator it)
{
    ONE_METRIC_COUNTER_DEC("fetches.concurrent");
    it->second.promise.setValue(it->second.wrote);
    fetchInfo_.erase(it);
}

void FetchManager::noteInProgress(const std::size_t inProgress)
{
    totalInProgress_ += inProgress;
    if (active_ && totalInProgress_ > FLAGS_max_incoming_buffered_size) {
        active_ = false;
        forEachPipeline([](auto pipeline) { pipeline->transportInactive(); });
    }
}

void FetchManager::noteDone(const std::size_t done)
{
    totalInProgress_ -= done;
    if (!active_ && totalInProgress_ < FLAGS_max_incoming_buffered_size) {
        active_ = true;
        forEachPipeline([](auto pipeline) { pipeline->transportActive(); });
    }
}

template <typename F>
void FetchManager::forEachPipeline(F &&f)
{
    for (auto &el : pipelines_)
        el.second->runInEventBaseThread(
            [ctx = el.first, f = std::forward<F>(f)] {
                f(static_cast<client::DataPipeline *>(ctx->getPipeline()));
            });
}

}  // namespace rtransfer
