#pragma once

#include "common.hpp"
#include "handleCache.hpp"
#include "storage.hpp"

#include <folly/FBString.h>
#include <folly/String.h>
#include <folly/futures/Future.h>
#include <folly/io/async/EventBase.h>
#include <monitoring/monitoring.h>
#include <wangle/channel/Handler.h>

#include <cstdint>
#include <unordered_map>

namespace rtransfer {

class FetchManager
    : public wangle::InboundHandler<std::unique_ptr<folly::IOBuf>,
          folly::Unit> {
public:
    FetchManager(StoragesMap &storages, HandleCache &handleCache,
        folly::EventBase *base);

    folly::Future<std::size_t> newFetch(std::uint64_t requestId,
        folly::fbstring destStorageId, folly::fbstring destFileId,
        std::uint64_t offset, std::size_t size,
        folly::Function<void(std::uint64_t, std::size_t)> notifyCb);

    void read(Context *ctx, std::unique_ptr<folly::IOBuf> buf) override;

    void transportActive(Context *ctx) override
    {
        auto pipelineBase = ctx->getPipeline()->getTransport()->getEventBase();
        base_->runInEventBaseThread([=] { pipelines_[ctx] = pipelineBase; });
    }

    void detachPipeline(Context *ctx) override
    {
        base_->runInEventBaseThread([=] { pipelines_.erase(ctx); });
    }

    void cancelFetch(std::uint64_t requestId);

    void setTotalSize(std::uint64_t reqId, std::size_t size);

private:
    struct FetchInfo {
        folly::fbstring storageId;
        folly::fbstring fileId;
        std::uint64_t offset;
        std::size_t size;
        folly::Promise<std::size_t> promise;
        std::size_t wrote;
        folly::Function<void(std::uint64_t, std::size_t)> notifyCb;
    };

    void doRead(std::unique_ptr<folly::IOBuf> buf);

    void cancelFetch(std::unordered_map<std::uint64_t, FetchInfo>::iterator it,
        folly::StringPiece reason);

    void fulfillPromise(
        std::unordered_map<std::uint64_t, FetchInfo>::iterator it);

    void noteInProgress(const std::size_t inProgress);
    void noteDone(const std::size_t done);
    template <typename F>
    void forEachPipeline(F &&f);

    StoragesMap &storages_;
    HandleCache &handleCache_;
    folly::EventBase *base_;
    std::unordered_map<std::uint64_t, FetchInfo> fetchInfo_;
    std::unordered_map<Context *, folly::EventBase *> pipelines_;
    std::size_t totalInProgress_ = 0;
    bool active_ = true;
};

}  // namespace rtransfer
