#pragma once

#include "openType.hpp"
#include "periodicHandler.hpp"

#include <folly/EvictingCacheMap.h>
#include <folly/FBString.h>
#include <folly/Hash.h>
#include <folly/Optional.h>
#include <folly/futures/SharedPromise.h>
#include <folly/io/async/EventBase.h>
#include <helpers/storageHelper.h>

#include <list>
#include <mutex>
#include <unordered_map>

namespace rtransfer {

class Storage;

class HandleCache {
    class TimerCallback : public folly::HHWheelTimer::Callback {
        friend class HandleCache;

    public:
        TimerCallback(HandleCache *cache);
        void timeoutExpired() noexcept override;
        void callbackCanceled() noexcept override;
        void schedule() noexcept;

    private:
        HandleCache *cache_;
    };

public:
    HandleCache(folly::EventBase *base);

    folly::Future<folly::IOBufQueue> read(Storage &storage, std::uint64_t reqId,
        folly::fbstring fileId, const std::uint64_t offset,
        const std::size_t size, std::uint8_t priority);

    folly::Future<folly::Unit> write(Storage &storage, folly::fbstring fileId,
        const std::uint64_t offset, folly::IOBufQueue buf,
        std::uint8_t priority);

private:
    using Key = std::tuple<Storage *, folly::fbstring, OpenType>;

    struct LRUNode {
        Key key;
        std::chrono::steady_clock::time_point addTimePoint;
    };

    struct Node {
        std::shared_ptr<folly::SharedPromise<one::helpers::FileHandlePtr>>
            promise;
        std::list<LRUNode>::iterator lruIt;
        std::size_t refCount = 0;
    };

    template <typename Cb>
    auto withHandle(Storage &storage, folly::fbstring fileId, OpenType openType,
        Cb &&cb) -> decltype(cb(std::declval<one::helpers::FileHandlePtr>()));

    folly::Future<one::helpers::FileHandlePtr> checkout(
        Storage &storage, folly::fbstring fileId, OpenType openType);

    void checkin(Storage &storage, folly::fbstring fileId, OpenType openType);

    void trigger();

    folly::EventBase *base_;
    std::unordered_map<Key, Node> handles_;
    std::list<LRUNode> lru_;

    std::unique_ptr<PeriodicHandler> periodicHandler_;
};

}  // namespace rtransfer
