#pragma once

#include <folly/FBVector.h>
#include <folly/Random.h>

#include <array>
#include <iostream>
#include <numeric>
#include <queue>
#include <tuple>
#include <unordered_map>

namespace rtransfer {

template <typename T>
class RandomizingPriorityQueue {
public:
    template <typename O>
    void push(std::uint8_t priority, O &&request)
    {
        queues_[priority].emplace_back(std::forward<O>(request));
        auto pit = nonemptyPrios_.begin();
        while (pit != nonemptyPrios_.end() && *pit < priority)
            ++pit;
        if (pit == nonemptyPrios_.end() || *pit != priority)
            nonemptyPrios_.insert(pit, priority);
    }

    bool empty() { return nonemptyPrios_.empty(); }

    std::pair<std::pair<std::size_t, std::size_t>, T *> top()
    {
        auto prioIdx = choosePriority();
        auto prio = nonemptyPrios_[prioIdx];
        auto idx = random_.rand64(queues_[prio].size());
        return {std::make_pair(prioIdx, idx), &queues_[prio][idx]};
    }

    void pop(std::pair<std::size_t, std::size_t> place)
    {
        auto prioIdx = place.first;
        auto prio = nonemptyPrios_[prioIdx];
        auto idx = place.second;
        std::swap(queues_[prio][idx], *queues_[prio].rbegin());
        queues_[prio].pop_back();
        if (queues_[prio].empty())
            nonemptyPrios_.erase(nonemptyPrios_.begin() + prioIdx);
    }

private:
    std::size_t choosePriority() const
    {
        auto sum = 1000;
        auto rand = random_.rand64(sum);

        std::size_t idx = 0;
        for (; idx < nonemptyPrios_.size() - 1 && rand > sum * 3 / 5; ++idx) {
            sum = sum * 3 / 5;
            rand -= sum;
        }
        return idx;
    }

    folly::Random random_;
    folly::fbvector<std::uint8_t> nonemptyPrios_;
    std::array<folly::fbvector<T>, std::numeric_limits<std::uint8_t>::max()>
        queues_;
};

}  // namespace rtransfer
