#include "reader.hpp"

#include "shaperMap.hpp"

namespace rtransfer {

Reader::Reader(StoragesMap &storages, HandleCache &handleCache,
    std::shared_ptr<ServerSideLink> serverLink, folly::EventBase *base)
    : storages_{storages}
    , handleCache_{handleCache}
    , serverLink_{std::move(serverLink)}
    , base_{base}
{
}

folly::Future<std::size_t> Reader::read(std::uint64_t reqId,
    folly::fbstring storageId, folly::fbstring fileId, std::uint64_t offset,
    std::size_t size, std::uint8_t priority)
{
    return via(base_).then(
        [=, storageId = std::move(storageId), fileId = std::move(fileId)] {
            return doRead(reqId, storageId, fileId, offset, size, priority);
        });
}

folly::Future<folly::Unit> Reader::cancel(
    std::uint64_t reqId, folly::fbstring storageId)
{
    return via(base_).then([=, storageId = std::move(storageId)] {
        return doCancel(reqId, storageId);
    });
}

folly::Future<std::size_t> Reader::doRead(std::uint64_t reqId,
    const folly::fbstring &storageId, const folly::fbstring &fileId,
    std::uint64_t offset, std::size_t size, std::uint8_t priority)
{
    auto it = storages_.find(storageId);
    if (it == storages_.cend())
        return folly::makeFuture<std::size_t>(
            folly::make_exception_wrapper<std::runtime_error>(
                "invalid storage"));

    VLOG(2) << "Reading " << reqId << " offset: " << offset
            << ", size: " << size;
    return handleCache_.read(*it->second, reqId, fileId, offset, size, priority)
        .via(base_)
        .then([this, reqId, priority, offset](folly::IOBufQueue buf) {
            VLOG(2) << "Choosing data conn for " << reqId
                    << ", offset: " << offset
                    << ", size: " << buf.chainLength();

            auto dataConn = serverLink_->randomDataConnection();
            std::uint64_t id = folly::Endian::big<std::uint64_t>(reqId);
            std::uint32_t size = buf.chainLength();
            std::uint32_t bigEndiansize =
                folly::Endian::big<std::uint32_t>(size);
            std::uint64_t bigEndianOffset =
                folly::Endian::big<std::uint64_t>(offset);

            auto header = folly::IOBuf::copyBuffer(&id, sizeof(id), 0,
                sizeof(priority) + sizeof(bigEndianOffset) +
                    sizeof(bigEndiansize));

            std::memcpy(header->writableTail(), &priority, sizeof(priority));
            header->append(sizeof(priority));

            std::memcpy(header->writableTail(), &bigEndianOffset,
                sizeof(bigEndianOffset));
            header->append(sizeof(bigEndianOffset));

            std::memcpy(
                header->writableTail(), &bigEndiansize, sizeof(bigEndiansize));
            header->append(sizeof(size));

            if (buf.chainLength() > 0)
                header->prependChain(buf.move());

            auto metric =
                "link." + serverLink_->getPeerAddr() + ".pre_write_buffer.size";
            ONE_METRIC_COUNTER_ADD(metric, size);

            return dataConn.via(base_).then(
                [=, header = std::move(header), metric = std::move(metric)](
                    wangle::DefaultPipeline::Ptr conn) mutable {
                    VLOG(2) << "Sending " << reqId << ", offset: " << offset
                            << ", size: " << size;

                    auto start = std::chrono::steady_clock::now();
                    return conn->write(std::move(header))
                        .via(base_)
                        .then([=, metric = std::move(metric)] {
                            auto delay =
                                std::chrono::duration_cast<
                                    std::chrono::nanoseconds>(
                                    std::chrono::steady_clock::now() - start) /
                                200;

                            if (transportCodel_.overloaded(delay) && shaperMap_)
                                shaperMap_->transportCongestion();

                            ONE_METRIC_COUNTER_SUB(metric, size);
                            VLOG(2)
                                << "Sent " << reqId << ", offset: " << offset
                                << ", size: " << size << ", took: "
                                << std::chrono::duration_cast<
                                       std::chrono::milliseconds>(delay)
                                       .count()
                                << "ms";

                            return size;
                        });
                });
        });
}

folly::Future<folly::Unit> Reader::doCancel(
    std::uint64_t reqId, folly::StringPiece storageId)
{
    auto it = storages_.find(storageId.fbstr());
    if (it == storages_.cend())
        return folly::makeFuture<folly::Unit>(
            folly::make_exception_wrapper<std::runtime_error>(
                "invalid storage"));

    return it->second->cancel(reqId);
}

}  // namespace rtransfer
