#pragma once

#include "common.hpp"
#include "proto/rtransfer_messages.pb.h"

#include <folly/FBString.h>
#include <folly/FBVector.h>
#include <folly/Random.h>
#include <folly/SpinLock.h>
#include <folly/concurrency/ConcurrentHashMap.h>
#include <glog/logging.h>
#include <wangle/channel/Handler.h>
#include <wangle/channel/Pipeline.h>

#include <sys/ioctl.h>

#include <algorithm>

namespace rtransfer {

class DataConn;

class ServerSideLink
    : public wangle::HandlerAdapter<std::unique_ptr<proto::LinkMessage>> {
public:
    ServerSideLink(std::string peerAddr, folly::EventBase *base)
        : peerAddr_{std::move(peerAddr)}
        , base_{base}
    {
    }

    std::string getPeerAddr() { return peerAddr_; }

    void attachPipeline(Context *ctx) override { ctrlCtx_ = ctx; }

    void addDataConnection(wangle::DefaultPipeline::Ptr pipeline)
    {
        std::lock_guard<folly::SpinLock> guard{lock_};
        dataConns_.emplace_back(pipeline);
    }

    void removeDataConnection(wangle::DefaultPipeline::Ptr pipeline)
    {
        std::lock_guard<folly::SpinLock> guard{lock_};
        dataConns_.emplace_back(pipeline);
        auto it = std::find(dataConns_.begin(), dataConns_.end(), pipeline);
        if (it != dataConns_.end()) {
            LOG(INFO) << "DATA conn down";
            dataConns_.erase(it);
        }
    }

    folly::Future<wangle::DefaultPipeline::Ptr> randomDataConnection()
    {
        std::lock_guard<folly::SpinLock> guard{lock_};
        static auto compare = [](const auto &a, const auto &b) {
            auto unsentData = [](auto pipeline) {
                auto socket = std::dynamic_pointer_cast<folly::AsyncSocket>(
                    pipeline->getTransport());
                CHECK(socket) << "transport unset or is not folly::AsyncSocket";

                int value;
                int error = ioctl(socket->getFd(), TIOCOUTQ, &value);
                PCHECK(error == 0) << "failed fo get socket send buffer size";

                return value;
            };

            return unsentData(a) < unsentData(b);
        };

        CHECK(!dataConns_.empty());
        return *std::min_element(dataConns_.begin(), dataConns_.end(), compare);
    }

    void readEOF(Context *ctx) override
    {
        for (auto conn : dataConns_)
            conn->getTransport()->getEventBase()->runInEventBaseThread(
                [conn] { conn->getTransport()->close(); });
        ctx->fireReadEOF();
    }

    void writeToPipeline(MsgPtr msg)
    {
        base_->runInEventBaseThread([=, msg = std::move(msg)]() mutable {
            return ctrlCtx_->fireRead(std::move(msg));
        });
    }

private:
    std::string peerAddr_;
    folly::EventBase *base_;
    Context *ctrlCtx_ = nullptr;
    folly::fbvector<wangle::DefaultPipeline::Ptr> dataConns_;
    folly::SpinLock lock_;
};  // namespace rtransfer

class DataConn : public wangle::HandlerAdapter<folly::IOBufQueue &,
                     std::unique_ptr<folly::IOBuf>> {
public:
    DataConn(std::shared_ptr<ServerSideLink> link)
        : link_{link}
    {
    }

    void attachPipeline(Context *ctx) override
    {
        link_->addDataConnection(getPipeline(ctx));
    }

    void readEOF(Context *ctx) override
    {
        link_->removeDataConnection(getPipeline(ctx));
        ctx->fireReadEOF();
    }

private:
    wangle::DefaultPipeline::Ptr getPipeline(Context *ctx)
    {
        return std::static_pointer_cast<wangle::DefaultPipeline>(
            ctx->getPipelineShared());
    }

    std::shared_ptr<ServerSideLink> link_;
};

class ServerSideLinkFactory {
public:
    ServerSideLinkFactory(
        folly::ConcurrentHashMap<folly::fbstring, std::weak_ptr<ServerSideLink>>
            &map)
        : map_{map}
    {
    }

    std::shared_ptr<ServerSideLink> operator()(
        folly::StringPiece ctrlId, std::string peerAddr, folly::EventBase *base)
    {
        auto link = std::make_shared<ServerSideLink>(std::move(peerAddr), base);
        map_.insert_or_assign(ctrlId.fbstr(), link);
        return link;
    }

    std::shared_ptr<DataConn> createDataConn(const folly::fbstring &ctrlId)
    {
        auto it = map_.find(ctrlId);
        if (it == map_.cend())
            return {};

        if (auto link = it->second.lock())
            return std::make_shared<DataConn>(link);

        map_.erase(ctrlId);
        return {};
    }

private:
    folly::ConcurrentHashMap<folly::fbstring, std::weak_ptr<ServerSideLink>>
        &map_;
};

}  // namespace rtransfer
