#include "service.hpp"

#include "serverSideLink.hpp"
#include "storage.hpp"

#include <monitoring/monitoring.h>

#include <folly/Bits.h>

namespace rtransfer {

folly::Future<Service::MsgPtr> Service::operator()(MsgPtr request)
{
    VLOG(1) << "Request from remote: " << request->DebugString();

    ONE_METRIC_COUNTER_INC("requests.remote.incoming.total");
    ONE_METRIC_COUNTER_INC("requests.remote.incoming.concurrent");
    auto reqId = request->req_id();
    return action(std::move(request))
        .onError([](const std::exception &e) { return error(e.what()); })
        .then([reqId = std::move(reqId)](MsgPtr resp) {
            if (resp) {
                resp->set_req_id(reqId);
                VLOG(1) << "Replying: " << resp->DebugString();
            }

            return resp;
        })
        .ensure([] {
            ONE_METRIC_COUNTER_DEC("requests.remote.incoming.concurrent");
        });
}

folly::Future<Service::MsgPtr> Service::action(MsgPtr request)
{
    switch (request->payload_case()) {
        case proto::LinkMessage::kFetch:
            ONE_METRIC_COUNTER_INC("requests.remote.incoming.fetch");
            return fetch(std::move(request));
        case proto::LinkMessage::kCongestion:
            ONE_METRIC_COUNTER_INC("requests.remote.incoming.congestion");
            return congestion(std::move(request));
        case proto::LinkMessage::kCancel:
            ONE_METRIC_COUNTER_INC("requests.remote.incoming.cancel");
            return cancel(std::move(request));
        default:
            ONE_METRIC_COUNTER_INC("requests.remote.incoming.bad_request");
            return error("invalid operation");
    }
}

folly::Future<Service::MsgPtr> Service::fetch(MsgPtr request)
{
    auto ask = std::make_unique<link_control::proto::Response>();
    auto authRequest = ask->mutable_auth_request();
    authRequest->set_transfer_data(request->fetch().transfer_data());
    authRequest->set_storage_id(request->fetch().src());
    authRequest->set_file_id(request->fetch().file_id());
    authRequest->set_provider_id(providerId_.data(), providerId_.size());
    return controlService_.ask(std::move(ask))
        .then([](link_control::proto::RequestPtr authResponse) {
            if (authResponse->payload_case() !=
                link_control::proto::Request::kAuthResponse)
                return folly::makeFuture<folly::Unit>(
                    folly::make_exception_wrapper<std::runtime_error>(
                        "internal authentication error"));

            if (!authResponse->auth_response().is_authorized())
                return folly::makeFuture<folly::Unit>(
                    folly::make_exception_wrapper<std::runtime_error>(
                        "unauthorized"));

            return folly::makeFuture();
        })
        .then([this, request = std::move(request)]() mutable {
            return shaperMap_.fetch(std::move(request));
        })
        .then([reqId = request->fetch().req_id()](std::size_t size) {
            VLOG(2) << "Sending back read result req_id: " << reqId
                    << ", size: " << size << " bytes";
            auto resp = std::make_unique<proto::LinkMessage>();
            resp->set_total_size(size);
            return resp;
        });
}

folly::Future<Service::MsgPtr> Service::cancel(MsgPtr request)
{
    return shaperMap_.cancel(std::move(request)).then([] {
        auto resp = std::make_unique<proto::LinkMessage>();
        resp->set_done(true);
        return folly::makeFuture<MsgPtr>(std::move(resp));
    });
}

folly::Future<Service::MsgPtr> Service::congestion(MsgPtr request)
{
    shaperMap_.congestion(std::move(request));
    return folly::makeFuture<Service::MsgPtr>(Service::MsgPtr{nullptr});
}

folly::Future<Service::MsgPtr> Service::error(folly::StringPiece description)
{
    auto msg = std::make_unique<proto::LinkMessage>();
    msg->set_error(description.data(), description.size());
    return folly::makeFuture(std::move(msg));
}

}  // namespace rtransfer
