#include "shaper.hpp"

#include <glog/logging.h>

DEFINE_uint64(shaper_max_increase, 10 * 1024 * 1024,
    "max shaper window increase per tick");
DEFINE_uint64(single_fetch_max_size, 10 * 1024 * 1024,
    "maximum size of single read block");
DEFINE_double(
    shaper_cubic_c, 0.001, "scaling factor for CUBIC window increase");
DEFINE_double(shaper_cubic_b, 0.2, "scaling factor for CUBIC window decrease");
DEFINE_uint64(shaper_cubic_initial_window, 10 * 1024 * 1024,
    "initial window size for CUBIC algorithm");

namespace {

std::uint8_t clampPriority(std::uint32_t priority)
{
    return priority <= std::numeric_limits<std::uint8_t>::max()
               ? static_cast<std::uint8_t>(priority)
               : std::numeric_limits<std::uint8_t>::max();
}

}  // namespace

namespace rtransfer {

Shaper::Shaper(folly::fbstring src, folly::fbstring dst, Reader &reader,
    folly::EventBase *base)
    : base_{base}
    , src_{std::move(src)}
    , dst_{std::move(dst)}
    , reader_{reader}
    , C{FLAGS_shaper_cubic_c}
    , B{FLAGS_shaper_cubic_b}
    , wMax_{FLAGS_shaper_cubic_initial_window}
    , wCubic_{FLAGS_shaper_cubic_initial_window}
{
}

folly::Future<std::size_t> Shaper::read(MsgPtr msg)
{
    return via(base_).then([this, msg = std::move(msg)]() mutable {
        return doRead(std::move(msg));
    });
}

folly::Future<std::size_t> Shaper::doRead(MsgPtr msg)
{
    VLOG(1) << "Storing req_id: " << msg->fetch().req_id() << " for shaping";

    auto priority = clampPriority(msg->fetch().priority());
    Request req{std::move(msg)};
    auto f = req.promise->getFuture();
    reqs_.emplace(req.fetch->fetch().req_id(), req.promise);
    delayedRequests_.push(priority, std::move(req));
    drainQueue();
    return f;
}

folly::Future<folly::Unit> Shaper::cancel(MsgPtr msg)
{
    return via(base_).then([this, msg = std::move(msg)]() mutable {
        return doCancel(std::move(msg));
    });
}

folly::Future<folly::Unit> Shaper::doCancel(MsgPtr msg)
{
    auto range = reqs_.equal_range(msg->cancel().req_id());
    for (auto it = range.first; it != range.second; ++it)
        it->second->setException(
            folly::make_exception_wrapper<std::runtime_error>(
                "Fetch cancelled by user"));

    reqs_.erase(range.first, range.second);
    return reader_.cancel(msg->cancel().req_id(), msg->cancel().src());
}

void Shaper::signalCongestion(std::chrono::steady_clock::time_point now)
{
    congestion_ = true;
}

void Shaper::drainQueue()
{
    while (sent_ < wCubic_ && !delayedRequests_.empty()) {
        auto res = delayedRequests_.top();
        if (res.second->promise->isFulfilled()) {
            delayedRequests_.pop(res.first);
            continue;
        }

        const std::size_t toRequest =
            std::min(wCubic_ - sent_, FLAGS_single_fetch_max_size);
        const std::size_t frontSize = res.second->fetch->fetch().size();

        if (frontSize <= toRequest) {
            fulfilWholeRequest(*res.second);
            delayedRequests_.pop(res.first);
            sent_ += frontSize;
        }
        else {
            fulfilSubRequest(*res.second, toRequest);
            sent_ += toRequest;
        }
    }
}

void Shaper::trigger(std::chrono::steady_clock::time_point now)
{
    // Trigger is called in parent EventBase thread, which currently is the same
    // as our base_
    if (congestion_) {
        congestion_ = false;
        wMax_ = wCubic_;
        epoch_ = now;
        K = calculateK(wMax_);
        recalculateWCubic(now);
        VLOG(2) << "Congestion wCubic: " << wCubic_;
    }
    else if (shouldIncreaseWindow()) {
        recalculateWCubic(now);
        VLOG(2) << "Increased wCubic: " << wCubic_;
    }
    else {
        // looks like we're good with our current window size, so just move
        // epoch forward to not increase the window too much later
        epoch_ += now - lastTrigger_;
    }

    lastTrigger_ = now;
    sent_ = 0;
    drainQueue();
}

void Shaper::recalculateWCubic(std::chrono::steady_clock::time_point now)
{
    const auto t =
        std::chrono::duration_cast<std::chrono::milliseconds>(now - epoch_);
    wCubic_ = std::min<double>(wCubic_ + FLAGS_shaper_max_increase,
        C * std::pow(t.count() - K, 3) + wMax_);
}

double Shaper::calculateK(std::size_t wMax)
{
    return std::pow(wMax * B / C, 1.0 / 3);
}

void Shaper::fulfilWholeRequest(Request &req)
{
    auto reqId = req.fetch->fetch().req_id();
    VLOG(2) << "fulfilWholeRequest req_id: " << reqId
            << ", offset: " << req.fetch->fetch().offset()
            << ", size: " << req.fetch->fetch().size();

    req.readFutures.emplace_back(readFromStorage(std::move(req.fetch)));
    collect(req.readFutures)
        .then([promise = req.promise](const std::vector<std::size_t> &vals) {
            std::size_t sum = 0;
            for (auto val : vals)
                sum += val;
            promise->setValue(sum);
        })
        .onError([promise = req.promise](folly::exception_wrapper ew) {
            promise->setException(std::move(ew));
        })
        .via(base_)
        .ensure([=] { reqs_.erase(reqId); });
}

void Shaper::fulfilSubRequest(Request &req, std::size_t size)
{
    VLOG(2) << "fulfilSubRequest req_id: " << req.fetch->fetch().req_id()
            << ", offset: " << req.fetch->fetch().offset()
            << ", size: " << size;

    auto subMsg = std::make_unique<proto::LinkMessage>(*req.fetch);
    subMsg->mutable_fetch()->set_size(size);
    auto origFetch = req.fetch->mutable_fetch();
    origFetch->set_size(origFetch->size() - size);
    origFetch->set_offset(origFetch->offset() + size);
    req.readFutures.emplace_back(readFromStorage(std::move(subMsg)));
}

folly::Future<std::size_t> Shaper::readFromStorage(MsgPtr req)
{
    auto &msg = req->fetch();
    auto priority = clampPriority(msg.priority());
    return reader_
        .read(msg.req_id(), msg.src(), msg.file_id(), msg.offset(), msg.size(),
            priority)
        .then([reqId = msg.req_id(), offset = msg.offset(), size = msg.size()](
                  folly::Try<std::size_t> res) {
            if (res.hasException()) {
                VLOG(1) << "Error fulfilling req_id: " << reqId
                        << ", offset: " << offset << ", size: " << size;
                return folly::makeFuture<std::size_t>(
                    std::move(res).exception());
            }
            VLOG(2) << "Successfuly read req_id: " << reqId
                    << ", offset: " << offset << ", size: " << size;
            return folly::makeFuture(res.value());
        });
}

}  // namespace rtransfer
