#pragma once

#include "common.hpp"
#include "priorityQueue.hpp"
#include "reader.hpp"

#include <folly/Hash.h>
#include <folly/futures/Future.h>
#include <gflags/gflags.h>
#include <wangle/channel/Handler.h>
#include <wangle/service/Service.h>
#include <boost/multi_index/hashed_index.hpp>
#include <boost/multi_index/key_extractors.hpp>
#include <boost/multi_index_container.hpp>

#include <atomic>
#include <queue>

namespace rtransfer {

struct Shaper {
    struct Request {
        Request(MsgPtr msg)
            : fetch{std::move(msg)}
        {
        }
        MsgPtr fetch;
        std::shared_ptr<folly::Promise<std::size_t>> promise{
            std::make_shared<folly::Promise<std::size_t>>()};
        folly::fbvector<folly::Future<std::size_t>> readFutures;
    };

public:
    Shaper(folly::fbstring src, folly::fbstring dst, Reader &reader,
        folly::EventBase *base);

    folly::fbstring src_;
    folly::fbstring dst_;

    folly::Future<std::size_t> read(MsgPtr msg);

    folly::Future<folly::Unit> cancel(MsgPtr msg);

    void signalCongestion(std::chrono::steady_clock::time_point now);

    void trigger(std::chrono::steady_clock::time_point now);

private:
    folly::Future<std::size_t> doRead(MsgPtr msg);

    folly::Future<folly::Unit> doCancel(MsgPtr msg);

    bool shouldIncreaseWindow() { return sent_ == wCubic_; }

    double calculateK(std::size_t wMax);

    void drainQueue();

    void fulfilWholeRequest(Request &req);

    void fulfilSubRequest(Request &req, std::size_t size);

    folly::Future<std::size_t> readFromStorage(MsgPtr msg);

    void recalculateWCubic(std::chrono::steady_clock::time_point now);

    const double C;
    const double B;
    folly::EventBase *base_;
    std::chrono::steady_clock::time_point epoch_{
        std::chrono::steady_clock::now()};
    std::chrono::steady_clock::time_point lastTrigger_{epoch_};
    std::size_t t_{0};  // milliseconds since last congestion event
    std::size_t wMax_;
    double K{calculateK(wMax_)};
    std::size_t wCubic_;
    std::size_t sent_{0};
    RandomizingPriorityQueue<Request> delayedRequests_;
    std::unordered_multimap<std::uint64_t,
        std::shared_ptr<folly::Promise<std::size_t>>>
        reqs_;
    Reader &reader_;
    std::atomic<bool> congestion_{false};
};

using ShaperPtr = std::shared_ptr<Shaper>;

}  // namespace rtransfer
