#pragma once

#include "bucketQueue.hpp"
#include "openType.hpp"

#include <folly/FBString.h>
#include <folly/Function.h>
#include <folly/Synchronized.h>
#include <folly/executors/Codel.h>
#include <folly/executors/IOThreadPoolExecutor.h>
#include <folly/executors/NamedThreadFactory.h>
#include <folly/io/IOBufQueue.h>
#include <helpers/storageHelper.h>

#include <atomic>
#include <chrono>
#include <queue>
#include <unordered_map>

namespace rtransfer {

class Storage {
public:
    Storage(folly::fbstring storageId,
        one::helpers::StorageHelperPtr storageHelper,
        folly::Function<void()> notifyReadCongestion,
        folly::Function<void()> notifyWriteCongestion);

    folly::Future<folly::IOBufQueue> read(one::helpers::FileHandlePtr,
        std::uint64_t reqId, const std::uint64_t offset, const std::size_t size,
        const std::uint8_t priority);

    folly::Future<folly::Unit> write(one::helpers::FileHandlePtr,
        const std::uint64_t offset, folly::IOBufQueue buf,
        std::uint8_t priority);

    folly::Future<one::helpers::FileHandlePtr> open(
        folly::fbstring fileId, OpenType openType);

    folly::Future<folly::Unit> cancel(std::uint64_t reqId);

private:
    struct ReadRequest {
        std::uint64_t reqId;
        folly::Promise<folly::IOBufQueue> promise;
        std::chrono::steady_clock::time_point requestTime;
        one::helpers::FileHandlePtr handle;
        std::uint64_t offset;
        std::size_t size;
        std::shared_ptr<bool> canceled;
    };

    struct WriteRequest {
        folly::Promise<folly::Unit> promise;
        std::chrono::steady_clock::time_point requestTime;
        one::helpers::FileHandlePtr handle;
        std::uint64_t offset;
        folly::IOBufQueue buf;
    };

    template <typename Ret, typename T>
    folly::Future<Ret> withFile(
        folly::fbstring fileId, OpenType openType, T &&cb);

    void handleRead(const std::size_t bucket);
    void handleWrite(const std::size_t bucket);
    void notifyReadCongestion(std::chrono::steady_clock::time_point now);
    void notifyWriteCongestion(std::chrono::steady_clock::time_point now);
    std::string m(const char *);

    template <typename Queue, typename NotifyCongestion, typename Cb>
    void usingBucket(Queue &queue, const std::size_t bucket,
        folly::Codel &codel, NotifyCongestion &&notifyCongestion, Cb &&cb);

    void logPop(WriteRequest &msg);
    void logPop(ReadRequest &msg);

    folly::fbstring storageIdHex_;
    one::helpers::StorageHelperPtr storageHelper_;
    folly::Function<void()> notifyReadCongestion_;
    folly::Function<void()> notifyWriteCongestion_;

    std::shared_ptr<folly::NamedThreadFactory> threadFactory_{
        std::make_shared<folly::NamedThreadFactory>("storage")};
    folly::IOThreadPoolExecutor executor_{1, threadFactory_};

    std::chrono::steady_clock::time_point lastReadCongestionNotif_;
    std::chrono::steady_clock::time_point lastWriteCongestionNotif_;

    constexpr static const std::size_t maxInProgress_{100};
    std::size_t currentReadsInProgress_{0};
    std::size_t currentWritesInProgress_{0};
    folly::Codel readCodel_;
    folly::Codel writeCodel_;

    template <typename T>
    struct ReqHash {
        std::size_t operator()(const T &req) const
        {
            return std::hash<void *>{}(static_cast<void *>(req.handle.get()));
        }
    };

    std::unordered_multimap<std::uint64_t, std::shared_ptr<bool>> cancelReq_;
    BucketQueue<ReadRequest, ReqHash<ReadRequest>> readQueue_;
    BucketQueue<WriteRequest, ReqHash<WriteRequest>> writeQueue_;
};

}  // namespace rtransfer
