#pragma once

#include "common.hpp"
#include "proto/control.pb.h"
#include "serverSideLink.hpp"
#include "storage.hpp"

#include <folly/FBString.h>
#include <folly/FBVector.h>
#include <folly/ThreadName.h>
#include <helpers/storageHelper.h>
#include <helpers/storageHelperCreator.h>
#include <asio/io_service.hpp>

#include <thread>

namespace rtransfer {

class StorageHelperFactory {
public:
    StorageHelperFactory(std::size_t threads, ClientLinksMap &clientLinks,
        ServerLinksMap &serverLinks)
        : helperIOService_{threads}
        , clientLinks_{clientLinks}
        , serverLinks_{serverLinks}
    {
        for (int i = 0; i < threads; ++i)
            helperWorkers_.emplace_back(std::thread{[this] {
                folly::setThreadName("helperIOS");
                helperIOService_.run();
            }});
    }

    ~StorageHelperFactory()
    {
        helperIOService_.stop();
        for (auto &thread : helperWorkers_)
            thread.join();
    }

    std::shared_ptr<Storage> operator()(
        const link_control::proto::CreateHelper &req)
    {
        std::unordered_map<folly::fbstring, folly::fbstring> params;
        for (auto &param : req.params())
            params.emplace(
                folly::fbstring{param.key()}, folly::fbstring{param.value()});

        auto helper = helperCreator_.getStorageHelper(
            req.name(), params, req.io_buffered());

        folly::Function<void()> notifyReadCongestion =
            [ this, storageId = req.storage_id() ]
        {
            for (auto linkIt = serverLinks_.cbegin();
                 linkIt != serverLinks_.cend(); ++linkIt) {
                if (auto link = linkIt->second.lock()) {
                    auto msg = std::make_unique<proto::LinkMessage>();
                    msg->mutable_congestion()->set_source(storageId);
                    link->writeToPipeline(std::move(msg));
                }
            }
        };

        folly::Function<void()> notifyWriteCongestion =
            [ this, storageId = req.storage_id() ]
        {
            for (auto linkIt = clientLinks_.cbegin();
                 linkIt != clientLinks_.cend(); ++linkIt) {
                auto link = linkIt->second;
                linkIt->second->congestion(storageId);
            }
        };

        return std::make_shared<Storage>(req.storage_id(), std::move(helper),
            std::move(notifyReadCongestion), std::move(notifyWriteCongestion));
    }

private:
    asio::io_service helperIOService_;
    ClientLinksMap &clientLinks_;
    ServerLinksMap &serverLinks_;

    asio::io_service::work helperWork_{helperIOService_};
    folly::fbvector<std::thread> helperWorkers_;

    one::helpers::StorageHelperCreator helperCreator_{helperIOService_,
        helperIOService_, helperIOService_, helperIOService_, helperIOService_,
        helperIOService_};
};

}  // namespace rtransfer
